/*
 * Copyright 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */


static int read_direct_item (struct inode * inode, loff_t offset, char * buf, int size)
{
  struct key key;
  int repeat, pos_in_item;
  struct path path;
  struct item_head * ih;
  char * local_buf, * p;
  int chars, left, read;
  struct buffer_head * bh;

  if (!INODE_OFFSET_IN_DIRECT (inode, offset + 1)) {
    reiserfs_warning ("vs-18010: read_direct_item: given offset (%d) is not stored in direct item. \
Inode's first direct byte %d\n", offset, inode->u.reiserfs_i.i_first_direct_byte);
    return 0;
  }

  local_buf = kmalloc (size, GFP_KERNEL);
  if (local_buf == 0) {
    return -ENOMEM;
  }
  p = local_buf;

  init_path (&path);

  copy_short_key (&key, INODE_PKEY(inode));
  key.k_offset = offset + 1;
  mark_key_direct (&key);

  left = size;
  read = 0;
  while (left) {
    if (search_for_position_by_key (inode->i_sb, &key, &path, &pos_in_item, &repeat) == POSITION_NOT_FOUND)
      break;

    bh = PATH_PLAST_BUFFER (&path);
    ih = B_N_PITEM_HEAD (bh, PATH_LAST_POSITION (&path));
    chars = ih->ih_item_len - pos_in_item;
    if (chars > left)
      chars = left;


    memcpy (p, B_I_PITEM (bh, ih) + pos_in_item, chars);
    key.k_offset += chars;
    read += chars;
    left -= chars;
  }

  copy_to_user (buf, local_buf, read);
  kfree (local_buf);
  pathrelse (&path);
  return read;
}


static ssize_t reiserfs_file_read (struct file * filp, char * buf, size_t count, loff_t * ppos)
{
  struct inode * inode = filp->f_dentry->d_inode;
  ssize_t read, to_be_read, offset;
  size_t size;
  int from_direct;

  /* wait direct_to_indirect and indirect_to_direct conversions to complete and lock them */
  lock_conversion (inode);

  offset = *ppos;
  size = inode->i_size;
  if (offset > size)
    to_be_read = 0;
  else
    to_be_read = size - offset;
  if (to_be_read > count)
    to_be_read = count;
  if (to_be_read <= 0)
    return 0;
  read = 0;

  if (!INODE_OFFSET_IN_DIRECT (inode, offset + 1)) {
    /* offset we start from is in unformatted node, use generic_file_read */
    read = generic_file_read (filp, buf, to_be_read, ppos);
    if (read == to_be_read	/* done */ ||
	read <= 0		/* read error */ ||
	!(inode->u.reiserfs_i.i_first_direct_byte - 1 < *ppos) /* current position is not in direct item */
	) {
      unlock_conversion (inode);
      return read;
    }

    /* it is very likely, that bytes stored in direct item are unread */
    from_direct = read % inode->i_sb->s_blocksize;
    read -= from_direct;
    to_be_read = from_direct;
    *ppos -= from_direct;

    /* we have to read direct item, set its offset */
    offset = inode->u.reiserfs_i.i_first_direct_byte - 1;
  }

  /* if there are bytes in direct items and we had to read them */
  if ((from_direct = read_direct_item (inode, offset, buf + read, to_be_read)) != to_be_read) {
    reiserfs_warning ("vs-18020: reiserfs_file_read: direct item/items of file do not have so many bytes (%d). Only %d. \
The rest filled with 0s.\nFile size %d.\nRead by generic_file_read %d\n", to_be_read, from_direct, size, read);
  }

  unlock_conversion (inode);



  /*printk ("reiserfs_file_read returns %d\n", read + to_be_read);*/
  *ppos += to_be_read;
 
  return read + to_be_read;
}

