/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */

#ifndef _LINUX_REISER_FS_SB
#define _LINUX_REISER_FS_SB
                                /* this struct goes on disk */
struct reiserfs_super_block
{
  __u32 s_block_count;			/* blocks count         */
  __u32 s_free_blocks;                  /* free blocks count    */
  __u32 s_root_block;           	/* root block number    */
  __u16 s_blocksize;                   	/* block size           */
  __u16 s_oid_maxsize;			/* max size of object id array, see get_objectid() commentary  */
  __u16 s_oid_cursize;			/* current size of object id array */
  __u16 s_state;                       	/* valid or error       */
  char s_magic[16];                     /* reiserfs magic string indicates that file system is reiserfs */
  __u16 s_tree_height;                  /* height of disk tree */
  __u16 s_bmap_nr;                      /* amount of bitmap blocks needed to address each block of file system */
  __u16 s_reserved;
};

#define SB_SIZE (sizeof(struct reiserfs_super_block))

#ifdef __KERNEL__

/*
 * reiserfs super block data in memory
 */


/* The preserve list stores blocknr's of blocks whose contents have in at least part been shifted by balancing to
   another node in RAM but perhaps not yet written to disk.  We do not attempt to track more than that a block has had
   its contents shifted since the last time that there were no dirty buffers that were suspected recipients (see
   mark_suspected_recipients).  The preserve list is constructed by kmalloc'ing a page at a time whenever the
   preserve_offset reaches near the end of a page.  Scheduling issues need to be remembered when growing the list. This
   list is append to the end and read from the beginning in its access.  This motivates the jar containing pointers to
   the end and the start. */

/* a jar holds a preserve list */
struct jar
{
  char * plist_first_page;	/* points to first memory block in a chain */
  char * plist_current_page;	/* points to memory block that contains pointer to free entry */
  unsigned long * plist_end;	/* points to free entry at end of preserve list */
};


struct reiserfs_sb_info
{
  struct buffer_head * s_sbh;                   /* Buffer containing the super block */
  struct reiserfs_super_block * s_rs;           /* Pointer to the super block in the buffer */
  struct buffer_head ** s_ap_true_bitmap;       /* array of buffers, holding block bitmap */
  struct buffer_head ** s_ap_cautious_bitmap;   /* the same as the above, but we do not set bit until new allocated
                                                   block reached disk, or more precisely, until the preserve list is
                                                   freed */
  unsigned short s_mount_state;                 /* reiserfs state (valid, invalid) */
  
  struct jar s_jam;                             /* holds the preserve list, see preserve.c */
  unsigned long read_lock;              /* locks fs for reads, which makes multiprocess large reads more efficient. We
                                           use bit 0 for the read lock.  This should be merged into the s_state field.
                                           Yura, if this read_locking code works well, when you find time, please do so.  */
  struct wait_queue * read_wait;                /* wait queue for reiserfs read lock */
  unsigned long s_suspected_recipient_count;    /* To support preserve lists we need to be able to
                                                   efficiently determine if there are any dirty
                                                   blocks that have not reached disk. This counter
                                                   serves that purpose.  It is incremented with
                                                   every dirtying of a block that might possibly
                                                   receive a shifted item from another buffer, and
                                                   decremented in unpreserve() which is called
                                                   whenever a buffer is written or freed.  See
                                                   unpreserve() and mark_buffer_dirty().  */

  void (*unpreserve)(struct super_block * s, struct buffer_head * bh); /* Vladimir, comment this.... -Hans */
  unsigned long s_mount_opt;

  /* to debug preserve list */
  char ** s_suspected_recipient_bitmap;
  char ** s_preserved_block_bitmap;
  int s_set_pb_bits;
  int s_set_sr_bits;

  /* session statistics */
  int s_kmallocs;
  int s_disk_reads;
  int s_disk_writes;
  int s_fix_nodes_repeats;
  int s_balance_count;
  int s_preserve_list_freeings;
  int s_unneeded_left_neighbor;
  int s_good_search_by_key_reada;
  int s_bmaps;
  int s_bmaps_without_search;
  int s_direct2indirect;
  int s_indirect2direct;

  /* for logging */
  int s_buffered_records;
  char * s_intermediate_log_buffer;
  struct buffer_head * s_log_bh;
  char * s_log_record;
};


#define NOTAIL 0  /* mount option -o notail */
#define NOPRESERVE 1 /*           -o nopreserve */ 
#define GENERICREAD 2 /*          -o genericread */
#define dont_have_tails(s) ((s)->u.reiserfs_sb.s_mount_opt & (1 << NOTAIL))
#define dont_preserve(s) ((s)->u.reiserfs_sb.s_mount_opt & (1 << NOPRESERVE))
#define use_genericread(s) ((s)->u.reiserfs_sb.s_mount_opt & (1 << GENERICREAD))

#define SB_PL_FIRST_PAGE(s) ((s)->u.reiserfs_sb.s_jam.plist_first_page)
#define SB_PL_CURRENT_PAGE(s) ((s)->u.reiserfs_sb.s_jam.plist_current_page)
#define SB_PL_END(s) ((s)->u.reiserfs_sb.s_jam.plist_end)


void __wait_on_inode (struct inode * inode);
extern struct list_head inode_in_use;
extern spinlock_t inode_lock;
inline int  test_and_wait_on_buffer(struct buffer_head * p_s_bh);
struct buffer_head * reiserfs_get_hash_table (kdev_t  n_dev, int n_block, int n_size, int * p_n_repeat);
struct buffer_head * reiserfs_getblk (kdev_t  n_dev, int n_block, int n_size, int * p_n_repeat);
void reiserfs_show_buffers (kdev_t dev);
void mark_suspected_recipients_dirty (kdev_t dev);
void invalidate_reiserfs_buffers (kdev_t dev);


#endif	/* __KERNEL__ */



#define SB_BUFFER_WITH_SB(s) ((s)->u.reiserfs_sb.s_sbh)
#define SB_DISK_SUPER_BLOCK(s) ((s)->u.reiserfs_sb.s_rs)
#define SB_BLOCK_COUNT(s) (SB_DISK_SUPER_BLOCK(s)->s_block_count)
#define SB_FREE_BLOCKS(s) (SB_DISK_SUPER_BLOCK(s)->s_free_blocks)
#define SB_REISERFS_MAGIC(s) (SB_DISK_SUPER_BLOCK(s)->s_magic)
#define SB_ROOT_BLOCK(s) (SB_DISK_SUPER_BLOCK(s)->s_root_block)
#define SB_TREE_HEIGHT(s) (SB_DISK_SUPER_BLOCK(s)->s_tree_height)
#define SB_REISERFS_STATE(s) (SB_DISK_SUPER_BLOCK(s)->s_state)
#define SB_BMAP_NR(s) (SB_DISK_SUPER_BLOCK(s)->s_bmap_nr)
#define SB_AP_TRUE_BITMAP(s) ((s)->u.reiserfs_sb.s_ap_true_bitmap)
#define SB_AP_CAUTIOUS_BITMAP(s) ((s)->u.reiserfs_sb.s_ap_cautious_bitmap)


#endif	/* _LINUX_REISER_FS_SB */






























